from app import app, login_manager, db
from views import views_init

app.logger.info("login manager init")
login_manager.init_app(app)
app.logger.info("db init")
db.create_all()
app.logger.info("views init")
views_init()

if __name__ == '__main__':
    app.run('0.0.0.0', '5000',  debug=True)


""" 
Purpose:    Main file about TPM service restful api.
            This project aimed to provide a TPM service for APPs.
            They can get TPM functions by calling APIs provided by this service. 

Attention:  This code should not be used in product directly, it just for demo.
            I use Flask-Restful API framwork(https://github.com/flask-restful/flask-restful)
            to create REST API for TPM service.
            Method call C binary file directly is not recommanded but used for now.
            TSS was writen in C, so python-C binding may be needed like libvirt-python.
            Or use ctypes...

About TPM:  TPM sorftware stack is ibm_tss.
            TPM2.0 simulator is ibm_sim.
            TSS can be replaced by other like MSR TSS or TCG TSS.
            TPM2.0 may provided by hardware, simulator, CUSE vtpm device.
            Whatever TSS or TPM is decoupled from this API.

TODO:       ACL and https, user mangement, requestParse

For test: 
            #/: curl -v -X PUT -H "Content-Type: application/json" -d '{"ic":"222"}' http://127.0.0.1:5000/tpm/tpm_hash
            return:
            {
                "hash_result": "9b871512327c09ce91dd649b3f96a63b7408ef267c8cc5710114e629730cb61f"
            }
"""

from flask import Flask
from flask_restful import Api
from flask_sqlalchemy import SQLAlchemy
from config import DevelopmentConfig
import flask_login
import logging

app = Flask("TPM_Service_APP")
app = Flask(__name__.split('.')[0])
api = Api(app, default_mediatype='application/json')
app.config.from_object(DevelopmentConfig)
db = SQLAlchemy(app)
login_manager = flask_login.LoginManager()
login_manager.login_message = "Login required"
app.logger.setLevel(logging.DEBUG)


import setuptools

setuptools.setup(
	name="TPM-Service",
	version="1.0.0",
	author="wenzt2",
	author_email="wenzt2@lenovo.com",
	description="Provide restful api service for TPM",
	packages=setuptools.find_packages(),
	platforms='linux',
	install_requires=[
		'Flask==1.1.1',
		'Flask-RESTful==0.3.7',
		'Flask-SQLAlchemy==2.4.0',
		'Flask-Login==0.4.1',
		'passlib==1.7.1',
	],
	classifiers=[
		'Development Status :: 1 - Beta',
		'Intended Audience :: Developers',
		'Programming Language :: Python',
		'Programming Language :: Python :: 2.7',
	],
)


from flask_restful import abort

def abort_if_todo_doesnt_exist(todo_id, todo_list):
    if todo_id not in todo_list:
        abort(404, message="TPM action {} doesn't exist".format(todo_id))

def abort_if_mimetype_not_json():
    if not request.json:
        abort(400, message="Content-Type must json")

def abort_if_input_doesnt_right(inputs, todo_id):
    if not tpm_input_validate.get(todo_id)(inputs):
        abort(405, message="Input params not right")

def validate_tpm_hash(inputs):
    if not 'ic' in input:
        return False

tpm_input_validate = {
    'validate_tpm_hash' : validate_tpm_hash
}



from flask import request
import flask_login
from app import app, db
from app import login_manager
from flask_restful import Resource
from passlib.hash import pbkdf2_sha256

class User(db.Model, flask_login.UserMixin):
    __tablename__ = 'users'
    id = db.Column(db.Integer, primary_key=True)
    #active means enable
    active = db.Column('is_active', db.Boolean(), nullable=False, server_default='1')
    # User Authentication fields
    username = db.Column(db.String(100, collation='NOCASE'), nullable=False, unique=True)
    password = db.Column(db.String(255), nullable=False, server_default='')
    roles = db.relationship('Role', secondary='user_roles')
    _is_authenticated = True

    @property
    def is_authenticated(self):
        return self._is_authenticated

    @is_authenticated.setter
    def is_authenticated(self, value):
        self._is_authenticated = value

    def hash_password(self, password):
        return pbkdf2_sha256.hash(password)

    def verify_password(self, password):
        return pbkdf2_sha256.verify(password, self.password)

class Role(db.Model):
    __tablename__ = 'roles'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(50), unique=True)

class UserRoles(db.Model):
    __tablename__ = 'user_roles'
    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer(), db.ForeignKey('users.id', ondelete='CASCADE')) 
    role_id = db.Column(db.Integer(), db.ForeignKey('roles.id', ondelete='CASCADE'))


def user_register(username, password):
    if not User.query.filter(User.username == username).first():
        app.logger.info("add user {0}".format(username))
        user = User(username = username)
        user.password = user.hash_password(password)
        db.session.add(user)
        db.session.commit()

#reloading a user from the session(client use cookies)
@login_manager.user_loader
def user_loader(user_id):
    print(user_id)
    user = User.query.get(user_id)
    if user and user.is_active:
        return user
    else:
        return None

#support for username:password auth; Basic Auth; api_key Auth
@login_manager.request_loader
def request_loader(request):
    username = request.form.get('username')
    if not User.query.filter(User.username == username).first():
        return

    user = User.query.filter(User.username == username).first()
    if user.verify_password(request.form['password']):
        user.is_authenticated = True
    else:
        user.is_authenticated = False
    return user

@login_manager.unauthorized_handler
def unauthorized_handler():
    return {
            "Message":"Login required", 
            "login":"tpm-api/login",
            "register":"tpm-api/register",
            "logout":"tpm-api/logout"
            }, 401

class User_Login(Resource):
    def post(self):
        #not safe use midle var and plaintext
        username = request.form.get('username')
        password = request.form.get('password')
        if not username or not password:
            abort(405, message="Input params not right") 
        user = User.query.filter(User.username == username).first()
        if not user:
            abort(400, message="User not exist")
        if user.verify_password(password):
            flask_login.login_user(user, False)
            return {"Message":"Login success"}, 200
        else:
            abort(405, message="Wrong password")    

class User_Logout(Resource):
    @flask_login.login_required
    def post(self):
        flask_login.logout_user()
        return "Logout success!"

class User_Register(Resource):
    def get(self):
        return {"Message":"User Register"}

    def post(self):
        username = request.form.get('username')
        password = request.form.get('password')#hash ?
        if User.query.filter_by(username=username).first() is not None: 
            abort(400, message="User exist")
        user_register(username, password) #except handle, user exist check
        return {"Message":"Register success!", "Username":username}, 201



from flask import request
from flask_restful import Resource, fields, marshal_with
import flask_login, tpm_utils
from tpm_actions import tpm_action, tpm_services
from usage import tpm_usage
from app import api
from auth import User_Login, User_Logout, User_Register
from app import app

api_resource_field = {
    'apiKey': fields.String,
    'apiUrl': fields.Url('service_ep', True)
}

service_resource_field = {
    'total': fields.Integer,
    'apis': fields.List(fields.Nested(api_resource_field))      
}

action_field = {
    'desc': fields.String,
    'apiKey': fields.String,
    'apiUrl': fields.Url('action_ep', True)
}
actions_resource_field = {
    'total': fields.Integer,
    'apis': fields.List(fields.Nested(action_field))
}

action_handle_field = {
    'desc': fields.String,
    'usage': fields.List(fields.String)
}

class Tpm_service_list(Resource):
    @marshal_with(service_resource_field)
    def get(self):
        data = {'total':len(tpm_services)}
        apis = []
        for v in tpm_services:
            apis.append({'apiKey': v, 'service_id': v})
        data['apis'] = apis
        return data

class Tpm_service(Resource):
    def get(self, service_id):
        tpm_utils.abort_if_todo_doesnt_exist(service_id, tpm_services)
        pass

class Tpm_action_list(Resource):
    @marshal_with(actions_resource_field)
    def get(self):
        data = {'total':len(tpm_action)}
        apis = []
        for v in tpm_action:
            apis.append({'apiKey': v, 'desc': tpm_usage.get(v).get('description'), 'action_id': v})
        data['apis'] = apis
        return data

class Tpm_action(Resource):
    @flask_login.login_required
    def get(self, action_id):
        tpm_utils.abort_if_todo_doesnt_exist(action_id, tpm_action)
        return tpm_usage.get(action_id)

    def put(self, action_id):
        tpm_utils.abort_if_todo_doesnt_exist(action_id, tpm_action)
        tpm_utils.abort_if_mimetype_not_json()
        input = request.get_json()
        #TODO: validate should replaced by RequestParser 
        tpm_utils.abort_if_input_doesnt_right(input, action_id)
        return tpm_action_handle.get(action_id)(input)

class Tpm_policy_list(Resource):
    def get(self):
        pass

class Tpm_policy(Resource):
    def get(self, todo_id):
        pass

    def put(self, todo_id):
        pass
def views_init():
    api.add_resource(Tpm_service_list, '/tpm-api')
    api.add_resource(Tpm_service,'/tpm-api/<service_id>', endpoint = 'service_ep')

    api.add_resource(Tpm_action_list, '/tpm-api/actions', endpoint = 'actions_ep')
    api.add_resource(Tpm_action, '/tpm-api/actions/<action_id>', endpoint = 'action_ep')

    api.add_resource(Tpm_policy_list, '/tpm-api/policys', endpoint = 'policys_ep')
    api.add_resource(Tpm_policy, '/tpm-api/policys/<policys_id>', endpoint = 'policy_ep')

    api.add_resource(User_Login, '/tpm-api/login', endpoint = 'login_ep')
    api.add_resource(User_Logout, '/tpm-api/logout', endpoint = 'logout_ep')
    api.add_resource(User_Register, '/tpm-api/register', endpoint = 'register_ep')


hash_usage = {
	'description' : 'hash gived string',
    'usage' : [
	    {
		    '[hi]':'hierarchy (e, o, p, n) (default null) e endorsement, o owner, p platform, n null',
		    '[halg]': '(sha1, sha256, sha384, sha512) (default sha256)',
		    'if': 'input file to be hashed',
		    'ic': 'data string to be hashed'
	    }
    ]
}

startup_usage = {
	'description' : 'Use PUT request to start up TPM device',
	'usage': [
		'No input needed'
	]
}

#tpm action usage
tpm_usage = {
    'tpm_startup' : startup_usage,
    'tpm_hash' : hash_usage
}


class DevelopmentConfig(object):
    SECRET_KEY = 'Change it dwfsfscvsvsefdafqfqdqdqwfe'
    SQLALCHEMY_DATABASE_URI = 'sqlite:///tpm_service.sqlite'
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    USER_APP_NAME = "TPM Service App"
    USER_ENABLE_EMAIL = False
    USER_ENABLE_USERNAME = True
    USER_REQUIRE_RETYPE_PASSWORD = False
    #REMEMBER_COOKIE_DURATION = 20 #seconds


"""
Put all functions here.
"""
import json
import subprocess
from app import app

tpm_tools_path = '/home/wenzt/tpm/ibm_tss/build/bin/'

def tpm_hash(input, halg='sha256'):

    cmd_line = [tpm_tools_path + 'tsshash']
    cmd_line.extend(['-ic', input['ic']])
    cmd_line.extend(['-halg', input.get('halg', 'sha256')])
    cmd_line.append('-ns')
    p = subprocess.Popen(cmd_line, stdout=subprocess.PIPE)
    p.wait()

    rv = p.stdout.read().strip()
    return {'hash_result' : rv}

def tpm_startup(input):
    cmd_line = [tpm_tools_path + 'tssstartup']
    p = subprocess.Popen(cmd_line, stdout=subprocess.PIPE)
    p.wait()

    rv = p.stdout.read().strip()
    return {'result' : rv}

#this dict define all methods that tpm/tpm_methods
tpm_action = [
    'tpm_startup',
    'tpm_hash'
]

#tpm action called by @GET @PUT @POST ...
tpm_action_handle = {
    'tpm_startup' : tpm_startup,
    'tpm_hash' : tpm_hash
}

#place here for now
tpm_services = [
    "actions",
    "policys",
    "register",
    "login",
    "logout",
]